<!DOCTYPE html>
<html lang="en">

<head>
  {% block head -%}
  <meta charset="UTF-8">

  <title>{% block title %}{{ title }}{% endblock %}</title>

  <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/css/bootstrap.min.css" integrity="sha384-1BmE4kWBq78iYhFldvKuhfTAU6auU8tT94WrHftjDbrCEXSU1oBoqyl2QvZ6jIW3" crossorigin="anonymous">
  <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap-icons@1.8.0/font/bootstrap-icons.css">
  <link rel="stylesheet" href="https://fonts.googleapis.com/icon?family=Material+Icons">
  {%- endblock %}
</head>

<body>
  <nav class="navbar mx-3" >
    <a class="navbar-brand" href="{{ url_for('browser.index_view') }}">Home</a>
    <a class="navbar-nav" href="{{ url_for('browser.chains_view') }}">Chains</a>
    {% block nav -%}
    {%- endblock %}
  </nav>

  {% block page_container -%}
  <div class="container-fluid">
    <div class="row"><div class="col">
      This space (un)intentionally left blank.
    </div></div>
  </div>
  {%- endblock %}

  {% block footer -%}
  <div class="footer text-center">
    Version {{ cc_version }} | <a href="https://cancelchain.org" class="link-dark">cancelchain.org</a>
  </div>
  {%- endblock %}

  {% block scripts -%}
  <script src="https://code.jquery.com/jquery-3.6.0.slim.min.js" integrity="sha256-u7e5khyithlIdTpu22PHhENmPcRdFiHRjhAuHcs05RI="crossorigin="anonymous"></script>

  <script src="https://cdn.jsdelivr.net/npm/bootstrap@5.1.3/dist/js/bootstrap.bundle.min.js" integrity="sha384-ka7Sk0Gln4gmtz2MlQnikT1wXgYsOg+OMhuP+IlRH9sENBO0LRn5q+8nbTov4+1p" crossorigin="anonymous"></script>
  {%- endblock %}
</body>

</html>
